"""util

This module contains some utility function used to help with the
implementation of the default strategies
"""

import errno
import os
import string
import unicodedata
from urllib.parse import urlparse


REPLACEMENT_CHAR = {'&': '-', ',': '.', ';': '-', '=': '_'}


def clean_filename(filename):
    """Return a sanitized filename (replace / strip out illegal characters)

    :param filename: string used for a filename
    :type filename: str

    :return: sanitized filename
    :rtype: str
    """

    return ''.join([
        c for c in unicodedata.normalize(
            'NFKD',
            ''.join([REPLACEMENT_CHAR.get(c, c) for c in filename])
        )
        if not unicodedata.combining(c) and c in '-_.() {0}{1}'.format(string.ascii_letters, string.digits)
    ])


def make_dirs(file_path):
    """Make the directories for a file path

    :param file_path: file path to be created if it doesn't exist
    :type file_path: str

    :return: None
    """

    try:
        os.makedirs(os.path.dirname(file_path))
    except OSError as exc:  # Guard against race condition
        if exc.errno != errno.EEXIST:
            raise  # pragma: no cover


def default_url_transform(url):
    """URL path segments are transformed into directories along a file path
    with the last path segment representing the filename.

    :param url: URL string
    :type url: str

    :return: file path
    :rtype: str
    """

    parsed_url = urlparse(url)

    path_segments = [
        clean_filename(path_segment) for path_segment in parsed_url.path.split('/')[1:]
        ]
    if not len(path_segments):
        path_segments = ['index']

    params = clean_filename(parsed_url.params)
    if len(params):
        params = '(' + params + ')'

    query = clean_filename(parsed_url.query)
    if len(query):
        query = '_' + query

    return os.path.sep.join([parsed_url.netloc] + path_segments) + params + query



"""Public API Functions

This module contains the public API functions.  The goal is to gain the
benefits of asynchronous HTTP requests while dropping one of these functions
into a synchronous code flow.  For basic usage, no prior knowledge of
asynchronous programming or asyncio is required.  Simply import aiodownload
and call the functions with some URLs.
"""

import asyncio

from . import AioDownloadBundle, AioDownload


def one(url_or_bundle, download=None):
    """Make one HTTP request and download it

    :param url_or_bundle: a URL string or bundle
    :type url_or_bundle: str or :class:`AioDownloadBundle`
    :param download: (optional) your own customized download object
    :type download: :class:`AioDownload`
    :return: a bundle
    :rtype: :class:`AioDownloadBundle`
    """

    return swarm([url_or_bundle], download=download)[0]


def swarm(iterable, download=None):
    """Make a swarm of requests and download them

    :param iterable: an iterable object (ex. list of URL strings)
    :type iterable: iterable object
    :param download: (optional) your own customized download object
    :type download: :class:`AioDownload`
    :return: a list of bundles
    :rtype: list
    """

    return [e for e in each(iterable, download=download)]


def each(iterable, url_map=None, download=None):
    """For each iterable object, map it to a URL and request asynchronously

    :param iterable: an iterable object (ex. list of objects)
    :type iterable: iterable object
    :param url_map: (optional) callable object mapping an object to a url or bundle
    :type url_map: callable object
    :param download: (optional) your own customized download object
    :type download: :class:`AioDownload`
    :return: generator
    """

    download = download or AioDownload()
    url_map = url_map or (lambda x: str(x))

    tasks = []
    for i in iterable:

        bundle = url_map(i)
        if not isinstance(bundle, AioDownloadBundle):
            bundle = AioDownloadBundle(bundle)

        if i != bundle.url:
            bundle.info = i

        tasks.append(download.loop.create_task(download.main(bundle)))

    for task_set in download.loop.run_until_complete(asyncio.wait(tasks)):
        for task in task_set:
            yield task.result()

    download.client.close()



"""aiodownload

This module contains the core classes.
"""

import aiohttp
import asyncio
import logging
import os

from .strategy import DownloadStrategy, Lenient

logger = logging.getLogger(__name__)

STATUS_ATTEMPT = 'Download attempted'
STATUS_CACHE = 'Cache hit'
STATUS_DONE = 'File written'
STATUS_FAIL = 'Download failed'
STATUS_INIT = 'Initialized'


class AioDownloadBundle:
    """A container class holding properties related to the URL.

    :class:`AioDownloadBundle`s get utilized by :class:`AioDownload`.  They
    hold information about the state of a bundle as they are processed.

    :param url: URL string (ex. https://www.google.com)
    :type url: str
    :param info: (optional) extra information that can be injected into the bundle
    :type info: dict
    :param params: (optional) params for a POST request
    :type params: dict
    """

    def __init__(self, url, info=None, params=None):

        self.attempts = 0  # value to be incremented by AioDownload when a request is attempted
        self.file_path = None  # determined by DownloadStrategy.url_transform
        self.info = info
        self.params = params
        self.url = url
        self._status_msg = STATUS_INIT  # set by AioDownload depending of the where it is in the flow of execution

    @property
    def status_msg(self):
        return '[URL: {}, File Path: {}, Attempts: {}, Status: {}]'.format(
            self.url, self.file_path, self.attempts, self._status_msg
        )


class AioDownload:
    """The core class responsible for the coordination of requests and downloads

    :param client: (optional) client session, a default is instantiated if not provided
    :type client: :class:`aiohttp.ClientSession`
    :param download_strategy: (optional) download strategy, a default is instantiated if not provided
    :type download_strategy: :class:`aiodownload.DownloadStrategy`
    :param request_strategy: (optional) request strategy, a :class:`Lenient` strategy is instantiated if not provided
    :type request_strategy: :class:`aiodownload.RequestStrategy`
    """

    def __init__(self, client=None, download_strategy=None, request_strategy=None):

        if not client:
            # Get the event loop and initialize a client session if not provided
            self.loop = asyncio.get_event_loop()
            self.client = aiohttp.ClientSession(loop=self.loop)
        else:
            # Or grab the event loop from the client session
            self.loop = client._loop
            self.client = client

        # Configuration objects managing download and request strategies
        self._download_strategy = download_strategy or DownloadStrategy()  # chunk_size, home, skip_cached
        self._request_strategy = request_strategy or Lenient()  # concurrent, max_attempts, timeout

        # Bounded semaphore guards how many requests can run concurrently
        self._main_semaphore = asyncio.BoundedSemaphore(self._request_strategy.concurrent)

    async def main(self, bundle):
        """Main entry point for task creation with an asyncio event loop.

        The number of concurrent requests is throttled using this async
        method.  Depending on the download strategy used, the method will call
        the request_and_download async method or immediately return the bundle
        indicating that the file came from cache as the file existed.

        :param bundle: bundle (generally one that has just been instantiated)
        :type bundle: :class:`aiodownload.AioDownloadBundle`

        :return: bundle with updated properties reflecting it's final state
        :rtype bundle: :class:`aiodownload.AioDownloadBundle`
        """

        with (await self._main_semaphore):

            bundle.file_path = self._download_strategy.get_file_path(bundle)
            file_exists = os.path.isfile(bundle.file_path)

            if not (file_exists and self._download_strategy.skip_cached):

                while bundle._status_msg in (STATUS_ATTEMPT, STATUS_INIT, ):

                    if bundle._status_msg == STATUS_ATTEMPT:
                        logger.info(bundle.status_msg)

                    sleep_time = self._request_strategy.get_sleep_time(bundle)
                    logger.debug('Sleeping {} seconds between requests'.format(sleep_time))
                    await asyncio.sleep(sleep_time)

                    bundle = await self.request_and_download(bundle)

            else:

                bundle._status_msg = STATUS_CACHE

            logger.info(bundle.status_msg)

        return bundle

    async def request_and_download(self, bundle):
        """Make an HTTP request and write it to disk.  Use the download and
        request strategies of the instance to implement how this is achieved.

        :param bundle: bundle with it's url and file_path set
        :type bundle: :class:`aiodownload.AioDownloadBundle`

        :return: bundle with updated properties reflecting success or failure
        :rtype bundle: :class:`aiodownload.AioDownloadBundle`
        """

        with aiohttp.Timeout(self._request_strategy.timeout):

            try:

                bundle.attempts += 1

                client_method = getattr(self.client, 'post' if bundle.params else 'get')
                async with client_method(bundle.url) as response:

                    try:

                        self._request_strategy.assert_response(response)
                        await self._download_strategy.on_success(response, bundle)
                        bundle._status_msg = STATUS_DONE

                    except AssertionError:

                        if self._request_strategy.retry(response):

                            if bundle.attempts >= self._request_strategy.max_attempts:

                                await self._download_strategy.on_fail(bundle)
                                bundle._status_msg = STATUS_FAIL

                            else:

                                bundle._status_msg = STATUS_ATTEMPT

                        else:

                            await self._download_strategy.on_fail(bundle)
                            bundle._status_msg = STATUS_FAIL

            except ValueError as err:

                bundle._status_msg = STATUS_FAIL
                logger.warning(' '.join([bundle.status_msg, str(err)]))

        return bundle



"""aiodownload API
"""

from .aiodownload import AioDownload, AioDownloadBundle
from .api import one, each, swarm
from .strategy import BackOff, DownloadStrategy, Lenient, RequestStrategy


__all__ = (
    'AioDownload',
    'AioDownloadBundle',
    'BackOff',
    'DownloadStrategy',
    'Lenient',
    'RequestStrategy',
    'one',
    'each',
    'swarm'
)



"""Strategies

This module contains base class and sensible default class implementations
for the request and download strategies used by :class:`AioDownload`
"""

import logging
import os

from .util import default_url_transform, make_dirs

logger = logging.getLogger(__name__)


class DownloadStrategy:
    """DownloadStrategy is an injection class for AioDownload.  The purpose is
    to control download options for AioDownload.

    :param chunk_size: the incremental chunk size to read from the response
    :type chunk_size: (optional) int
    :param home: the base file path to use for writing response content to file
    :type honme: (optional) str
    :param skip_cached: indicates whether existing written files should be skipped
    :type skip_cached: bool
    """

    def __init__(self, chunk_size=65536, home=None, skip_cached=False):
        self.chunk_size = chunk_size
        self.home = home or os.getcwd()
        self.skip_cached = skip_cached

    async def on_fail(self, bundle):
        """Write an empty file

        :param bundle: bundle (file_path should exist)
        :type bundle: :class:`AioDownloadBundle`

        :return: None
        """

        make_dirs(bundle.file_path)
        open(bundle.file_path, 'wb+').close()

    async def on_success(self, response, bundle):
        """Write the response to the file path indicated in the bundle

        :param response: successful response from an HTTP response
        :type response: :class:`aiohttp.ClientResponse`
        :param bundle: bundle (file_path should exist)
        :type bundle: :class:`AioDownloadBundle`

        :return: None
        """

        make_dirs(bundle.file_path)

        with open(bundle.file_path, 'wb+') as f:
            while True:
                chunk = await response.content.read(self.chunk_size)
                if not chunk:
                    break
                f.write(chunk)

    def get_file_path(self, bundle):
        """Get the file path for the bundle

        :param bundle: bundle (generally, it's file_path should be None)
        :type bundle: :class:`AioDownloadBundle`

        :return: full file_path for the bundle (via the default URL transformation)
        :rtype: str
        """

        return os.path.sep.join([
            self.home,
            default_url_transform(bundle.url)       # transforms the URL into a relative file path
        ])


class RequestStrategy:
    """RequestStrategy is an injection class for AioDownload.  The purpose is
    to control how AioDownload performs requests and retries requests.

    :param concurrent: the number of concurrent asynchronous HTTP requests to maintain
    :type concurrent: (optional) int
    :param max_attempts: maximum number of attempts before giving up
    :type max_attempts: int
    :param time_out: timeout for the client session
    :type time_out: int
    """

    def __init__(self, concurrent=2, max_attempts=0, timeout=60):
        self.concurrent = concurrent
        self.max_attempts = max_attempts
        self.timeout = timeout

    def assert_response(self, response):
        """Assertion for the response

        :param response: response from an HTTP response
        :type response: :class:`aiohttp.ClientResponse`
        :return: None or AssertionError
        """

        assert response.status < 400

    def retry(self, response):
        """Retry the HTTP request based on the response

        :param response: response from an HTTP response
        :type response: :class:`aiohttp.ClientResponse`

        :return:
        :rtype: bool
        """

        return False

    def get_sleep_time(self, bundle):
        """Returns how much time the bundle should sleep based on it's properties

        :param bundle: bundle
        :type bundle: :class:`AioDownloadBundle`

        :return: sleep time
        :rtype: int
        """

        return -1


class Lenient(RequestStrategy):
    """Lenient request strategy designed for an average web server.  Try five
    times with a minute between each retry.
    """

    def __init__(self, max_attempts=5):
        RequestStrategy.__init__(self, max_attempts)

    def retry(self, response):
        """Retry any unsuccessful HTTP response except a 404 (if they say
        it's not there, let's believe them)
        """

        return False if response.status == 404 else True

    def get_sleep_time(self, bundle):

        # Retry pattern: 0.25, 60, 60, 60, 60
        return 0.25 if bundle.attempts == 0 else 60


class BackOff(RequestStrategy):
    """Back Off request strategy designed for APIs and web servers that can
    be hammered a little harder.  Exponentially back away if a failed
    response is returned.
    """

    def __init__(self, max_attempts=10):
        RequestStrategy.__init__(self, max_attempts)

    def get_sleep_time(self, bundle):

        # Retry pattern: 0.25, 0.5, 1, 2, 4, 8, 16, 32, 60, 60
        return min(2**(bundle.attempts - 2), 60)
